# -.- coding: utf-8 -.-
#
# GNOME Activity Journal
#
# Copyright © 2009-2010 Seif Lotfy <seif@lotfy.com>
# Copyright © 2009-2010 Siegfried Gevatter <siegfried@gevatter.com>
# Copyright © 2007 Alex Graveley <alex@beatniksoftware.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os.path
import gobject
import urllib
import os
import gtk
import gnome.ui
import pango
import datetime


from zeitgeist.datamodel import Event, Subject, Interpretation, Manifestation

class IconFactory:
    '''
    Icon lookup swiss-army knife (from menutreemodel.py)
    '''

    def load_icon_from_path(self, icon_path, icon_size = None):
        if os.path.isfile(icon_path):
            try:
                if icon_size:
                    # constrain height, not width
                    return gtk.gdk.pixbuf_new_from_file_at_size(icon_path, -1, int(icon_size))
                else:
                    return gtk.gdk.pixbuf_new_from_file(icon_path)
            except:
                pass
        return None

    def load_icon_from_data_dirs(self, icon_value, icon_size = None):
        data_dirs = None
        if os.environ.has_key("XDG_DATA_DIRS"):
            data_dirs = os.environ["XDG_DATA_DIRS"]
        if not data_dirs:
            data_dirs = "/usr/local/share/:/usr/share/"

        for data_dir in data_dirs.split(":"):
            retval = self.load_icon_from_path(os.path.join(data_dir, "pixmaps", icon_value),
                                              icon_size)
            if retval:
                return retval
            
            retval = self.load_icon_from_path(os.path.join(data_dir, "icons", icon_value),
                                              icon_size)
            if retval:
                return retval

        return None

    def scale_to_bounded(self, icon, size):
        if icon:
            if icon.get_height() > size:
                return icon.scale_simple(size * icon.get_width() / icon.get_height(),
                                         size,
                                         gtk.gdk.INTERP_BILINEAR)
            else:
                return icon
        else:
            return None

    def load_icon(self, icon_value, icon_size, force_size = True):
        assert icon_value, "No icon to load!"

        if isinstance(icon_size, gtk.IconSize):
            icon_size = gtk.icon_size_lookup(icon_size)[0]
            force_size = True

        if isinstance(icon_value, gtk.gdk.Pixbuf):
            if force_size:
                return self.scale_to_bounded(icon_value, icon_size)
            return icon_value

        if os.path.isabs(icon_value):
            icon = self.load_icon_from_path(icon_value, icon_size)
            if icon:
                if force_size:
                    return self.scale_to_bounded(icon, icon_size)
                return icon
            icon_name = os.path.basename(icon_value)
        else:
            icon_name = icon_value
    
        if icon_name.endswith(".png"):
            icon_name = icon_name[:-len(".png")]
        elif icon_name.endswith(".xpm"):
            icon_name = icon_name[:-len(".xpm")]
        elif icon_name.endswith(".svg"):
            icon_name = icon_name[:-len(".svg")]
    
        icon = None
        info = icon_theme.lookup_icon(icon_name, icon_size, gtk.ICON_LOOKUP_USE_BUILTIN)
        if info:
            if icon_name.startswith("gtk-"):
                # NOTE: IconInfo/IconTheme.load_icon leaks a ref to the icon, so
                #       load it manually.
                # NOTE: The bindings are also broken for Gtk's builtin pixbufs:
                #       IconInfo.get_builtin_pixbuf always returns None.
                icon = info.load_icon()
            elif info.get_filename():
                icon = self.load_icon_from_path(info.get_filename())
        else:
            icon = self.load_icon_from_data_dirs(icon_value, icon_size) # Fallback

        if icon and force_size:
            return self.scale_to_bounded(icon, icon_size)
        return icon

    def load_image(self, icon_value, icon_size, force_size = True):
        pixbuf = self.load_icon(icon_value, icon_size, force_size)
        img = gtk.Image()
        img.set_from_pixbuf(pixbuf)
        img.show()
        return img

    def make_icon_frame(self, thumb, icon_size = None, blend = False):
        border = 1

        mythumb = gtk.gdk.Pixbuf(thumb.get_colorspace(),
                                 True,
                                 thumb.get_bits_per_sample(),
                                 thumb.get_width(),
                                 thumb.get_height())
        mythumb.fill(0x00000080) # black, 50% transparent
        if blend:
            thumb.composite(mythumb, 0, 0,
                            thumb.get_width(), thumb.get_height(),
                            0, 0,
                            1.0, 1.0,
                            gtk.gdk.INTERP_NEAREST,
                            128)
        thumb.copy_area(border, border,
                        thumb.get_width() - (border * 2), thumb.get_height() - (border * 2),
                        mythumb,
                        border, border)
        return mythumb

    def transparentize(self, pixbuf, percent):
        pixbuf = pixbuf.add_alpha(False, '0', '0', '0')
        for row in pixbuf.get_pixels_array():
            for pix in row:
                pix[3] = min(int(pix[3]), 255 - (percent * 0.01 * 255))
        return pixbuf

    def colorshift(self, pixbuf, shift):
        pixbuf = pixbuf.copy()
        for row in pixbuf.get_pixels_array():
            for pix in row:
                pix[0] = min(255, int(pix[0]) + shift)
                pix[1] = min(255, int(pix[1]) + shift)
                pix[2] = min(255, int(pix[2]) + shift)
        return pixbuf

    def greyscale(self, pixbuf):
        pixbuf = pixbuf.copy()
        for row in pixbuf.get_pixels_array():
            for pix in row:
                pix[0] = pix[1] = pix[2] = (int(pix[0]) + int(pix[1]) + int(pix[2])) / 3
        return pixbuf


class Thumbnailer:
    def __init__(self, uri, mimetype):
        self.uri = uri or ""
        self.mimetype = mimetype or ""
        self.cached_icon = None
        self.cached_timestamp = None
        self.cached_size = None

    def get_icon(self, icon_size, timestamp = 0):
        if not self.cached_icon or \
               icon_size != self.cached_size or \
               timestamp != self.cached_timestamp:
            self.cached_icon = self._lookup_or_make_thumb(icon_size, timestamp)
            self.cached_size = icon_size
            self.cached_timestamp = timestamp
        return self.cached_icon

    def _lookup_or_make_thumb(self, icon_size, timestamp):
        ### FIXME: Trash URIs crash older libgnomeui's icon lookup, so we avoid it
        if self.uri.startswith("trash:"):
            return self._get_trash_icon()

        icon_name, icon_type = \
                   gnome.ui.icon_lookup(icon_theme, thumb_factory, self.uri, self.mimetype, 0)
        try:
            if icon_type == gnome.ui.ICON_LOOKUP_RESULT_FLAGS_THUMBNAIL or \
                   thumb_factory.has_valid_failed_thumbnail(self.uri, timestamp):
                # Use existing thumbnail
                thumb = icon_factory.load_icon(icon_name, icon_size)
            elif self._is_local_uri(self.uri):
                # Generate a thumbnail for local files only
                print " *** Calling generate_thumbnail for", self.uri
                thumb = thumb_factory.generate_thumbnail(self.uri, self.mimetype)
                thumb_factory.save_thumbnail(thumb, self.uri, timestamp)

            if thumb:
                # Fixup the thumbnail a bit
                thumb = self._nicer_dimensions(thumb)
                thumb = icon_factory.make_icon_frame(thumb, icon_size)

                return thumb
        except:
            pass

        # Fallback to mime-type icon on failure
        return icon_factory.load_icon(icon_name, icon_size)

    def _get_trash_icon(self, uri):
        trash_icon = "gnome-fs-trash-full"
        import gimmie_trash
        if gimmie_trash.trash_monitor.is_empty():
            trash_icon = "gnome-fs-trash-empty"
        return icon_factory.load_icon(trash_icon, icon_size)

    def _is_local_uri(self, uri):
        # NOTE: gnomevfs.URI.is_local seems to hang for some URIs (e.g. ssh
        #       or http).  So look in a list of local schemes which comes
        #       directly from gnome_vfs_uri_is_local_scheme.
        scheme, path = urllib.splittype(self.get_uri() or "")
        return not scheme or scheme in ("file", "help", "ghelp", "gnome-help", "trash",
                                        "man", "info", "hardware", "search", "pipe",
                                        "gnome-trash")

    def _nicer_dimensions(self, icon):
        ### Constrain thumb dimensions to 1:1.2
        if float(icon.get_height()) / float(icon.get_width()) > 1.3:
            return icon.subpixbuf(0, 0,
                                  icon.get_width(), int(icon.get_width() * 1.3))
        return icon



def get_category_icon(icon, size=24):
    try:
        return icon_theme.load_icon (icon, size, 0)
    except:
        return None


icon_factory = IconFactory()
icon_theme = gtk.icon_theme_get_default()
thumb_factory = gnome.ui.ThumbnailFactory("normal")

ITEM_CACHE = {}



class Applet(gtk.HBox):
    def __init__(self):
        gtk.HBox.__init__(self)
        
        #self.systopic = SystemTopic()
        #self.apptopic = ApplicationsTopic()
        self.libtopic = LibraryTopic()
        #self.ppltopic = PeopleTopic()
        
        self.topics = [self.libtopic]
        for topic in self.topics:
            self.pack_start(topic)
            topic.connect("toggled",self._toggle_view)
        
        self.topic_window = TopicsWindow(self.topics)
        self.show_all()
        self.topic_window.hide()
        
    def _toggle_view(self, widget):
        #print widget
        if widget.get_active():
            self._position_window(widget)
            self.topic_window.show_all()
        else:
            self.topic_window.hide_all()
        
    def _position_window(self, topic):
        root_x, root_y = topic.window.get_origin()
        root_height = topic.allocation.height
        x = root_x
        y = root_y + root_height
        self.topic_window.move(x, y)
        

class Topic(gtk.ToggleButton):     
    def __init__(self, label):   
        gtk.ToggleButton.__init__(self, label)
        self.label = label
        self.set_relief(gtk.RELIEF_NONE)
        self.set_focus_on_click(False)
        self.categories = []
        

class LibraryTopic(Topic):
    def __init__(self):
        Topic.__init__(self, "Library")
        from datasources.library import library
        self.categories += library.get_toplevel_sources()
    
        
class TopicsWindow(gtk.Window):
    def __init__(self, topics):
        gtk.Window.__init__(self)
        self.topics = topics
        self.set_size_request(640, 400)
        self.set_decorated(False)
        self.set_border_width(3)
        
        self.notebook = gtk.Notebook()
        self.notebook.set_show_border(False)
        self.notebook.set_show_tabs(False)
        for topic in self.topics:
            topicview = TopicView(topic.categories)
            self.notebook.append_page(topicview, gtk.Label(topic.label))
            topic.connect("clicked", self._switch_window)
        self.add(self.notebook)
        self.connect("focus-out-event", self._hide_window)
        self.show_all()
        
    def _hide_window(self, window, event):
        #print window, event
        pass
        
    def _switch_window(self, widget):
        self.notebook.set_page(self.topics.index(widget))
        
class TopicView(gtk.VBox):
    def __init__(self, categories):
        gtk.VBox.__init__(self)
        self.categories = categories
        self.hbox = gtk.HBox()
        self.view = CategoryView()
        self.sidebar = Sidebar()
        self.sidebar.set_view(self.view)
        self.hbox.pack_start(self.sidebar, False, False)
        self.vbox = gtk.VBox()
        
        self.pack_start(self.hbox)
        scroll = gtk.ScrolledWindow()
        scroll.add_with_viewport(self.view)
        scroll.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
        scroll.get_children()[0].set_shadow_type(gtk.SHADOW_NONE)

        topbar = TimeBar()
        ev = gtk.EventBox()
        ev.add(topbar)
        self.vbox.pack_start(ev, False, False)
        self.vbox.pack_start(gtk.HSeparator(), False, False)
        self.vbox.pack_start(scroll)   
        
        scroll =gtk.ScrolledWindow()
        scroll.add_with_viewport(self.vbox)        
        scroll.set_policy(gtk.POLICY_NEVER, gtk.POLICY_NEVER)
        self.hbox.pack_start(scroll)
        
        topbar.connect("zoom-changed", self._zoom_changed)
        topbar.connect("sorting-changed", self._sort_changed)
        
        for cat in categories:
            if isinstance(cat, str):
                self.sidebar.pack_start(gtk.HSeparator(), False, False)
            else:
                self.sidebar.append_category(cat)
        self.sidebar.connect("clicked", self.set_view)
        
        self.sidebar.buttons[0].set_active(True)
        
        def change_style(widget, style):
            pass

        self.view.connect("style-set", change_style)
        
    def _sort_changed(self, sender, sorting):
        self.sidebar.set_sorting(sorting)

    def _zoom_changed(self, sender, zoom_level):
        self.sidebar.set_zoom_level(zoom_level)
        
            
    def set_view(self, button, items):
        if len(items) > 0:
            item = items[0]
        
        self.view.clear()
        for item in items:
            self.view.add_item(item)
        #self.view.update_size(item.icon_size)
                
                    
class Sidebar(gtk.VBox): 
    __gsignals__ = {
        "clicked" : (gobject.SIGNAL_RUN_FIRST,
                    gobject.TYPE_NONE,
                    (gobject.TYPE_PYOBJECT,))
    }   
    def __init__(self):
        gtk.VBox.__init__(self)
        self.busy = False
        self.buttons = []
        
    def set_view(self, view):
        self.view = view
        
    def set_zoom_level(self, zoom_level):
        for button in self.buttons:
            button.category.set_zoom_level(zoom_level)
            if button.get_active():
                button.emit("toggled")
                
    def set_sorting(self, sorting):
        for button in self.buttons:
            button.category.set_sorting(sorting)
            if button.get_active():
                button.emit("toggled")
        
    def append_category(self, cat):
        
        def callback(button):
            def callback2(items):
                self.emit("clicked", items)
            button.get_items(callback2)
        
        btn = CategoryButton(cat)
        self.pack_start(btn, False, False)
        self.buttons.append(btn)
        btn.connect_after("toggled", callback)
        btn.connect("toggled", self.untoggle_all)
        
    def untoggle_all(self, btn):
        if not self.busy:
            self.busy = True
            for b in self.buttons:
                if not b == btn: 
                    b.set_active(False)
            self.busy = False
            
        
class CategoryButton(gtk.ToggleButton):
    def __init__(self, category):
        gtk.ToggleButton.__init__(self)
        self.category = category       
        icon = icon_factory.load_icon(self.category.icon, 24)
        label = gtk.Label(category.title)
        hbox = gtk.HBox()
        hbox.pack_start(gtk.image_new_from_pixbuf(icon), False, False)
        hbox.pack_start(label)
        self.add(hbox)
        self.set_relief(gtk.RELIEF_NONE)
        self.set_border_width(1)
        self.set_focus_on_click(False)
        #self.set_alignment(0.0, 0.5)
        
    def get_items(self, callback):
        self.category.get_items(callback)
        
class CategoryView(gtk.IconView):
    def __init__(self):
        gtk.IconView.__init__(self)
        self.model = gtk.ListStore(gtk.gdk.Pixbuf, str, gobject.TYPE_PYOBJECT)
        self.set_margin(12)
        self.use_cells = isinstance(self, gtk.CellLayout)
        self.set_spacing(4)
        self.set_selection_mode(gtk.SELECTION_MULTIPLE)
        self.icon_cell = gtk.CellRendererPixbuf()
        self.icon_cell.set_property("xalign", 0.5)
        self.icon_cell.set_property("yalign", 1.0)
        self.icon_cell.set_fixed_size(128, -1)
        self.pack_start(self.icon_cell, expand=False)
        self.add_attribute(self.icon_cell, "pixbuf", 0)

        self.text_cell = gtk.CellRendererText()
        #self.text_cell.set_property("wrap-mode", pango.WRAP_WORD_CHAR)
        self.text_cell.set_property("yalign", 0.0)
        self.text_cell.set_property("xalign", 0.5)
        self.text_cell.set_property("ellipsize", pango.ELLIPSIZE_END)
        self.pack_start(self.text_cell, expand=True)
        self.add_attribute(self.text_cell, "markup", 1)
        #self.set_item_width(150)
        self.set_model(self.model)
        #self.set_orientation(gtk.ORIENTATION_HORIZONTAL)            
        #self.text_cell.set_property("wrap-width", 100)

        
    def clear(self):
        self.model.clear()
    
    def add_item(self, item):
        item = ITEM_CACHE[item]
        icon = item.get_icon()
        text = "<span>" + item.title + "</span>\n" + "<span size='small'>"+ item.comment + "</span>"
        text = text.replace("%20", " ")
        text = text.replace("%5B", "["); text = text.replace("%5D", "]")
        self.model.append([icon, text, item])
        
    def update_size(self, size):
        self._update_sizing(size)
        
    def _update_sizing(self, icon_width):
        pass
        #self.icon_cell.set_fixed_size(icon_width, -1) # Reset icon sizing
            #wrap_width = 150 - icon_width - 4
            #self.text_cell.set_property("wrap-width", wrap_width)
        
class TimeBar(gtk.EventBox):
    __gsignals__ = {
        "zoom-changed" : (gobject.SIGNAL_RUN_LAST,
                          gobject.TYPE_NONE,
                          (gobject.TYPE_INT,)),
        "sorting-changed" : (gobject.SIGNAL_RUN_LAST,
                          gobject.TYPE_NONE,
                          (gobject.TYPE_INT,)),
        }

    _zoomlist = [[_("Today"), 1],
                 [_("This week"), 7],
                 [_("Last 2 weeks"), 14],
                 [_("This month"), 30],
                 [_("Last 2 months"), 60],
                 [_("Last 3 months"), 90],
                 [_("Last 6 months"), 180]]
    _default_zoom = 7

    def __init__(self):
        gtk.EventBox.__init__(self)
        self.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse("white"))
        
        self.sorting = 2
        
        box = gtk.HBox(False, 0)
        box.show()
        self.add(box)

        self._add_sparkline(box)

        self.out_btn = gtk.Button()
        self.out_btn.add(gtk.image_new_from_stock(gtk.STOCK_ADD, gtk.ICON_SIZE_SMALL_TOOLBAR))
        self.out_btn.set_relief(gtk.RELIEF_NONE)
        self.out_btn.connect("clicked", lambda btn: self._zoom_clicked(1))
        box.pack_end(self.out_btn, False, False)
        
        self.in_btn = gtk.Button()
        self.in_btn.add(gtk.image_new_from_stock(gtk.STOCK_REMOVE, gtk.ICON_SIZE_SMALL_TOOLBAR))
        self.in_btn.set_relief(gtk.RELIEF_NONE)
        self.in_btn.connect("clicked", lambda btn: self._zoom_clicked(-1))
        box.pack_end(self.in_btn, False, False)
        
        self.label = gtk.Label()
        self.label.set_width_chars(max([len(x[0]) for x in self._zoomlist]))
        self.label.set_alignment(1.0, 0.5)
        box.pack_end(self.label, False, False, 6)

        box.show_all()

        self.set_zoom_level(self._default_zoom)
        
        self.combobox = gtk.combo_box_entry_new_text()
        label = gtk.Label(" Sort By: ")
        hbox = gtk.HBox()
        hbox.pack_start(label, False, False)
        hbox.pack_start(self.combobox, False, False)
        box.pack_start(hbox, False, False)
        
        self.combobox.append_text("Recency")
        self.combobox.append_text("Popularity")
        self.combobox.set_active(0)
        
        self.combobox.connect("changed", self._sorting_changed)
        
    def _sorting_changed(self, widget):
        #print widget
        self.sorting = widget.get_active()
        
        if self.sorting == 0:
            self.sorting = 2
        else:
            self.sorting = 4
            
        self.emit("sorting-changed", self.sorting)

    def _add_sparkline(self, box):
        self.spark = None
        
        ### Disabled due to bugginess for 0.2.8
        #self.spark = SparkLine(self.get_zoom_level_list(floor=0))
        #self.spark.connect("zoom-changed", lambda w, num_days: self.set_zoom_level(num_days))

        #align = gtk.Alignment(0.5, 0.5, 1.0, 1.0)
        #align.set_property("top-padding", 4)
        #align.add(self.spark)
        #box.pack_start(align, True, True, 6)

    def get_zoom_level(self):
        return self.zoom_level

    def get_zoom_level_list(self, floor = _default_zoom):
        return [days for name, days in self._zoomlist if days >= floor]

    def _zoom_clicked(self, dir):
        for i in range(len(self._zoomlist)):
            label, days = self._zoomlist[i]
            if days == self.zoom_level:
                try:
                    label, days = self._zoomlist[i + dir]
                except IndexError:
                    label, days = self._zoomlist[0]
                self.set_zoom_level(days, label)
                break

    def set_zoom_level(self, zoom_level, label=None):
        self.zoom_level = zoom_level

        if not label:
            for _label, days in self._zoomlist:
                if days == zoom_level:
                    label = _label
        self.label.set_text(label)
        if self.spark:
            self.spark.set_zoom_level(zoom_level)
        self.emit("zoom-changed", zoom_level)

    def set_items(self, items):
        dates = [x.get_timestamp() for x in items if x.get_timestamp() != 0]
        dates = [datetime.date.fromtimestamp(x) for x in dates]
        if self.spark:
            self.spark.set_dates(dates)


lib_applet = Applet()


